import configparser
import os
import tempfile
import gzip
from unittest import TestCase
from pathlib import Path
import gridfs
import datetime
import json
import zipfile
import shutil
from pymongo import errors, MongoClient
from bson import ObjectId, DBRef
from bson.json_util import dumps
from edman import Config, Convert, DB, File, Search
from edman.exceptions import EdmanDbProcessError


class TestFile(TestCase):
    db = None
    db_server_connect = False
    test_ini = []
    client = None

    @classmethod
    def setUpClass(cls):

        # 設定読み込み
        settings = configparser.ConfigParser()
        settings.read(Path.cwd() / 'ini' / 'test_db.ini')
        cls.test_ini = dict(settings.items('DB'))
        cls.test_ini['port'] = int(cls.test_ini['port'])

        # DB作成のため、pymongoから接続
        cls.client = MongoClient(cls.test_ini['host'], cls.test_ini['port'])

        # 接続確認
        try:
            cls.client.admin.command('hello')
            cls.db_server_connect = True
            print('Use DB.')
        except errors.ConnectionFailure:
            print('Do not use DB.')

        if cls.db_server_connect:
            # adminで認証
            cls.client = MongoClient(
                username=cls.test_ini['admin_user'],
                password=cls.test_ini['admin_password'])
            # DB作成
            cls.client[cls.test_ini['db']].command(
                "createUser",
                cls.test_ini['user'],
                pwd=cls.test_ini['password'],
                roles=[
                    {
                        'role': 'dbOwner',
                        'db': cls.test_ini['db'],
                    },
                ],
            )
            # edmanのDB接続オブジェクト作成
            con = {
                'host': cls.test_ini['host'],
                'port': cls.test_ini['port'],
                'user': cls.test_ini['user'],
                'password': cls.test_ini['password'],
                'database': cls.test_ini['db'],
                'options': [f"authSource={cls.test_ini['db']}"]
            }
            cls.db = DB(con)
            cls.testdb = cls.db.get_db

    @classmethod
    def tearDownClass(cls):
        if cls.db_server_connect:
            # cls.clientはpymongo経由でDB削除
            # cls.testdb.dbはedman側の接続オブジェクト経由でユーザ(自分自身)の削除
            cls.client.drop_database(cls.test_ini['db'])
            cls.testdb.command("dropUser", cls.test_ini['user'])

    def setUp(self):

        if self.db_server_connect:
            self.file = File(self.db.get_db)
        else:
            self.file = File()
        self.config = Config()

    # def tearDown(self):
    #     pass

    @staticmethod
    def make_txt_files(dir_path, name='file_dl_list', suffix='.txt',
                       text='test', qty=1):
        # 添付ファイル用テキストファイル作成
        p = Path(dir_path)
        if qty == 1:
            filename = name + suffix
            save_path = p / filename
            with save_path.open('w') as f:
                f.write(text)
        else:
            for i in range(qty):
                filename = name + str(i) + suffix
                save_path = p / filename
                with save_path.open('w') as f:
                    f.write(text + str(i))
        return sorted(p.glob(name + '*' + suffix))

    def test_file_gen(self):

        # 正常系
        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='insert_file_ref_test',
                                        qty=2)
            expected = []
            for file_path in files:
                with file_path.open() as f:
                    expected.append((file_path.name, f.read()))

            # ファイル読み込み、テスト
            actual = []
            # files = tuple(sorted(p.glob('gen_test*.txt')))
            for idx, f in enumerate(self.file.file_gen(tuple(files))):
                filename, filedata = f
                actual.append((filename, filedata.decode()))
            self.assertListEqual(sorted(actual), sorted(expected))

    def test_file_list_attachment(self):

        # _ed_fileがなかった場合
        doc = {'name': 'NSX', 'ddd': 'aaa'}
        files_oid = [ObjectId(), ObjectId()]
        actual = self.file.file_list_attachment(doc, files_oid)
        expected = {'name': 'NSX', 'ddd': 'aaa', self.config.file: files_oid}
        self.assertDictEqual(expected, actual)

        # _ed_fileがすでに存在する場合
        oid1 = ObjectId()
        oid2 = ObjectId()

        files_oid = [oid1, oid2]
        doc = {'abc': '123', self.config.file: files_oid}
        at_files_oid = [ObjectId(), ObjectId()]
        actual = self.file.file_list_attachment(doc, at_files_oid)

        files_oid2 = [oid1, oid2]
        expected = {'abc': '123', self.config.file: files_oid2 + at_files_oid}

        self.assertDictEqual(actual, expected)

    def test_file_list_replace(self):
        # _ed_fileがある場合
        files_oid = [ObjectId(), ObjectId()]
        doc = {'name': 'NSX', self.config.file: files_oid}
        rep_files_oid = [ObjectId(), ObjectId()]
        actual = self.file.file_list_replace(doc, rep_files_oid)
        expected = {'name': 'NSX', self.config.file: rep_files_oid}
        self.assertDictEqual(expected, actual)

        # 空リストの場合
        doc = {'name': 'NSX', self.config.file: [ObjectId(), ObjectId()]}
        rep_files_oid = []
        actual = self.file.file_list_replace(doc, rep_files_oid)
        expected = {'name': 'NSX'}
        self.assertDictEqual(expected, actual)

        # _ed_fileがない場合(例外になる)
        doc = {'name': 'NSX'}
        rep_files_oid = [ObjectId(), ObjectId()]
        with self.assertRaises(ValueError):
            _ = self.file.file_list_replace(doc, rep_files_oid)

    def test_fs_delete(self):
        if not self.db_server_connect:
            return

        # 正常系
        with tempfile.TemporaryDirectory() as tmp_dir:
            fs_oids = []
            for i in tuple(
                    self.make_txt_files(tmp_dir, name='delete_test', qty=2)):
                with i.open('rb') as f:
                    self.fs = gridfs.GridFS(self.testdb)
                    fs_oids.append(
                        self.fs.put(f.read(), filename=str(i.name)))

            self.file.fs_delete(fs_oids)
            for i in fs_oids:
                with self.subTest(i=i):
                    self.assertFalse(self.fs.exists(i))

    def test_get_file_names(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='file_names', qty=2)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in files:
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))

            # DocをDBに入れる
            doc = {
                'test1': {
                    'test2': 'name',
                    'test3': {
                        'test4': [
                            {
                                'test5': 'moon',
                                self.config.file: files_oid,
                            },
                            {'test6': 'star'}
                        ]
                    }
                }
            }
            insert_result = self.testdb['structure_emb'].insert_one(doc)

            # 正常系テスト(embのみ.refとembの判断はget_file_ref()で行っているので割愛)
            oid = insert_result.inserted_id
            query = ['test1', 'test3', 'test4', '0']
            actual = self.file.get_file_names('structure_emb', oid, 'emb',
                                              query)
            expected = dict(zip(files_oid, [i.name for i in files]))
            self.assertDictEqual(actual, expected)

            # 正常系 ファイルがなかった場合 →空のリストを出力
            doc = {
                'test1': {
                    'test2': 'name2',
                    'test3': {
                        'test4': [
                            {'test5': 'moon2'},
                            {'test6': 'star2'}
                        ]
                    }
                }
            }
            insert_result = self.testdb['structure_ref'].insert_one(doc)
            oid = insert_result.inserted_id
            actual = self.file.get_file_names('structure_ref', oid, 'ref')
            self.assertDictEqual(actual, {})

            # 異常系
            # ドキュメントがなかった場合 →EdmanDbProcessErrorを補足
            oid = ObjectId()
            with self.assertRaises(EdmanDbProcessError):
                _ = self.file.get_file_names('structure_ref', oid, 'ref')

    def test__get_file_ref(self):
        # 正常系(ref)
        expected = [ObjectId(), ObjectId(), ObjectId()]
        doc = {
            'name': 'test1',
            self.config.file: expected,
            self.config.parent: DBRef('test0', ObjectId()),
            self.config.child: [DBRef('test2', ObjectId()),
                                DBRef('test3', ObjectId())]
        }
        structure = 'ref'
        actual = self.file.get_file_ref(doc, structure)
        self.assertListEqual(expected, actual)

        # 正常系(emb)
        expected = [ObjectId(), ObjectId(), ObjectId()]
        doc = {
            'test1': {
                'test2': 'name',
                'test3': {
                    'test4': [
                        {
                            'test5': 'moon',
                            self.config.file: expected,
                        },
                        {'test6': 'star'}
                    ]
                }
            }
        }
        structure = 'emb'
        query = ['test1', 'test3', 'test4', '0']
        actual = self.file.get_file_ref(doc, structure, query)
        self.assertListEqual(expected, actual)

    def test_delete(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dl_dir:

            # embの添付ファイル削除テスト
            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            dl_files = self.make_txt_files(tmp_dl_dir, name='file_delete')
            for filename in dl_files:
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))

            # DocをDBに入れる
            doc = {
                'test1': {
                    'test2': 'name',
                    'test3': {
                        'test4': [
                            {
                                'test5': 'moon',
                                self.config.file: files_oid,
                            },
                            {'test6': 'star'}
                        ]
                    }
                }
            }
            insert_result = self.testdb['structure_emb'].insert_one(doc)
            oid = str(insert_result.inserted_id)
            query = ['test1', 'test3', 'test4', '0']

            result = self.file.delete(files_oid[0], 'structure_emb', oid,
                                      'emb', query)
            self.assertTrue(result)
            # ファイルが消えたか検証
            self.assertFalse(self.fs.exists(files_oid[0]))

            # refの添付ファイル削除テスト
            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in dl_files:
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))

            # docをDBに入れる
            doc = {'name': 'test1',
                   self.config.file: files_oid,
                   self.config.parent: ObjectId(),
                   self.config.child: DBRef('child_col', ObjectId())}

            insert_result = self.testdb['structure_ref'].insert_one(doc)
            oid = str(insert_result.inserted_id)

            result = self.file.delete(files_oid[0], 'structure_ref', oid,
                                      'ref')
            self.assertTrue(result)
            # ファイルが消えたか検証
            self.assertFalse(self.fs.exists(files_oid[0]))

    def test__get_emb_files_list(self):

        # 正常系
        file_list = [ObjectId(), ObjectId()]

        doc = {
            "_id": ObjectId(),
            "position": "top",
            "username": "ryde",
            "bool": True,
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "float_val": 4453.456
                },
                {
                    "maker": "HONDA",
                    "carname": "NSX",
                    "power": 280,
                    "float_val": 321.56,
                    "structure_3_1": [
                        {
                            "filename": "test1.txt",
                            "name": "添付ファイル1"
                        },
                        {
                            "filename": "test2.txt",
                            "name": "添付ファイル2"
                        }
                    ],
                    "structure_3_2": [
                        {
                            "filename": "test3.txt",
                            "name": "添付ファイル3",
                            "structure_4": {
                                "filename": "test4.txt",
                                "name": "添付ファイル4",
                                "structure_5": [
                                    {
                                        "url": "example2.com",
                                        "name": "テストURL2"
                                    },
                                    {
                                        "url": "example3.com",
                                        "name": "テストURL3"
                                    }
                                ]
                            },
                            "structure_6": [
                                {
                                    "url": "example_x.com",
                                    "name": "テストURL_x"
                                },
                                {
                                    "url": "example_y.com",
                                    "name": "テストURL_y",
                                    self.config.file: file_list,
                                }
                            ],
                            "structure_5": {
                                "url": "example.com",
                                "name": "テストURL1",
                                "structure_5": {
                                    "url": "example4.com",
                                    "name": "テストURL4"
                                }
                            }
                        }
                    ]
                }
            ]
        }

        query = ['structure_2', '1', 'structure_3_2', '0', 'structure_6', '1']
        actual = self.file._get_emb_files_list(doc, query)

        expected = file_list
        self.assertEqual(expected, actual)

    def test_zipped_json(self):

        # # 正常系
        s = {"document": [{"test1": "ABC"}, {"test2": 12345}]}
        raw_json = dumps(s, ensure_ascii=False, indent=4)
        encoded_json = raw_json.encode('utf-8')
        filename = 'jsonfile'
        ac_z_path = self.file.zipped_json(encoded_json, filename)

        # ac_z_pathを解凍して中身のファイル名がzipped_filepathと同じか調べる
        with tempfile.TemporaryDirectory() as tmpdir:
            with zipfile.ZipFile(ac_z_path, 'r') as inputFile:
                inputFile.extractall(tmpdir)
            p = Path(tmpdir)
            # 解凍されたディレクトリ名を取得する
            # print(list(p.glob('**/*.json')))
            ps = list(p.glob('**/*.json'))[0]
            actual = ps.stem

        expected = filename
        self.assertEqual(expected, actual)
        os.unlink(ac_z_path)

    def test_zipped_json2(self):

        # # 正常系
        s = {"document": [{"test1": "ABC"}, {"test2": 12345}]}
        raw_json = dumps(s, ensure_ascii=False, indent=4)
        encoded_json = raw_json.encode('utf-8')
        filename = 'jsonfile'

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir)
            ac_z_path = self.file.zipped_json_path(encoded_json, filename, p)

            # 圧縮前のjsonファイルを削除する
            first_file = filename + '.json'
            first = p / first_file
            first.unlink()

            # ac_z_pathを解凍して中身のファイル名がzipped_filepathと同じか調べる
            with zipfile.ZipFile(ac_z_path, 'r') as inputFile:
                inputFile.extractall(tmpdir)

            # 解凍されたディレクトリ名を取得する
            ps = list(p.glob('**/*.json'))[0]
            actual = ps.stem

        expected = filename
        self.assertEqual(expected, actual)

    def test_zipped_contents(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            # 添付ファイル用テキストファイル作成

            test_var = 'test'
            name = 'file_dl_list.txt'
            input_filename = os.path.splitext(os.path.basename(name))[0]
            file_dl_list = self.make_txt_files(tmp_dl_dir, name=input_filename,
                                               text=test_var)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in file_dl_list:
                metadata = {'filename': filename.name, 'compress': None}
                with filename.open('rb') as f:
                    files_oid.append(self.fs.put(f.read(), **metadata))

            # docをDBに入れる
            parent_id = ObjectId()
            doc_id = ObjectId()
            child_id = ObjectId()
            parent_col = 'parent_col'
            doc_col = 'doc_col'
            child_col = 'child_col'
            insert_docs = [
                {
                    'col': parent_col,
                    'doc': {
                        '_id': parent_id,
                        'name': 'parent',
                        Config.child: [DBRef(doc_col, doc_id)]
                    },
                },
                {
                    'col': doc_col,
                    'doc': {
                        '_id': doc_id,
                        'name': 'doc',
                        Config.file: files_oid,
                        Config.parent: DBRef(parent_col, parent_id),
                        Config.child: [DBRef(child_col, child_id)]
                    }
                },
                {
                    'col': child_col,
                    'doc': {
                        '_id': child_id,
                        'name': 'child',
                        Config.parent: DBRef(doc_col, doc_id),
                    }
                }]
            result = {}
            for i in insert_docs:
                insert_result = self.testdb[i['col']].insert_one(i['doc'])
                result.update({i['col']: insert_result.inserted_id})

            # # ドキュメントを取得する
            search = Search(self.db)
            exclusion = ['_id', Config.file]
            docs = search.find(doc_col, {'_id': result[doc_col]}, 1, 1,
                               exclusion)
            result_with_filepath, downloads = self.file.get_fileref_and_generate_dl_list(
                docs, "_ed_attachment")
            res = search.process_data_derived_from_mongodb(
                result_with_filepath)

        # 実行
        raw_json = dumps(res, ensure_ascii=False, indent=4)
        encoded_json = raw_json.encode('utf-8')
        json_tree_file_name = 'json_tree'
        zip_filepath = self.file.zipped_contents(downloads,
                                                 json_tree_file_name,
                                                 encoded_json)
        with tempfile.TemporaryDirectory() as tmp:
            # 解凍
            shutil.unpack_archive(zip_filepath, tmp)
            # 添付ファイルを読み込んでテキストの中身を抽出
            path_lists = [tmp, str(list(downloads.keys())[0]), name]
            with open(os.path.join(*path_lists)) as f:
                attached_s = f.read()
            path_lists = [tmp, json_tree_file_name + '.json']
            json_file_path = os.path.join(*path_lists)
            with open(json_file_path) as jf:
                j_data = json.load(jf)
            # 外に出すデータを組み立て(zip内部のjsonデータ、上記の添付ファイルデータ)
            actual_data = {'json_data': j_data, 'attached_data': attached_s}
        # テスト
        self.assertDictEqual(res, actual_data['json_data'])
        self.assertEqual(test_var, actual_data['attached_data'])

        # テスト用zipファイル削除
        if os.path.exists(zip_filepath):
            os.remove(zip_filepath)

    # def test_zipped_contents_path(self):
    #     if not self.db_server_connect:
    #         return
    #
    #     # TODO テストを書いている途中
    #
    #     with tempfile.TemporaryDirectory() as tmp_dl_dir:
    #         # 添付ファイル用テキストファイル作成
    #
    #         test_var = 'test'
    #         name = 'file_dl_list.txt'
    #         input_filename = os.path.splitext(os.path.basename(name))[0]
    #         file_dl_list = self.make_txt_files(tmp_dl_dir, name=input_filename,
    #                                            text=test_var)
    #
    #         # ファイル読み込み、ファイルをgridfsに入れる
    #         files_oid = []
    #         self.fs = gridfs.GridFS(self.testdb)
    #         for filename in file_dl_list:
    #             metadata = {'filename': filename.name, 'compress': None}
    #             with filename.open('rb') as f:
    #                 files_oid.append(self.fs.put(f.read(), **metadata))
    #
    #         # docをDBに入れる
    #         parent_id = ObjectId()
    #         doc_id = ObjectId()
    #         child_id = ObjectId()
    #         parent_col = 'parent_col'
    #         doc_col = 'doc_col'
    #         child_col = 'child_col'
    #         insert_docs = [
    #             {
    #                 'col': parent_col,
    #                 'doc': {
    #                     '_id': parent_id,
    #                     'name': 'parent',
    #                     Config.child: [DBRef(doc_col, doc_id)]
    #                 },
    #             },
    #             {
    #                 'col': doc_col,
    #                 'doc': {
    #                     '_id': doc_id,
    #                     'name': 'doc',
    #                     Config.file: files_oid,
    #                     Config.parent: DBRef(parent_col, parent_id),
    #                     Config.child: [DBRef(child_col, child_id)]
    #                 }
    #             },
    #             {
    #                 'col': child_col,
    #                 'doc': {
    #                     '_id': child_id,
    #                     'name': 'child',
    #                     Config.parent: DBRef(doc_col, doc_id),
    #                 }
    #             }]
    #         result = {}
    #         for i in insert_docs:
    #             insert_result = self.testdb[i['col']].insert_one(i['doc'])
    #             result.update({i['col']: insert_result.inserted_id})
    #
    #         # # ドキュメントを取得する
    #         search = Search(self.db)
    #         exclusion = ['_id', Config.file]
    #         docs = search.find(doc_col, {'_id': result[doc_col]}, 1, 1,
    #                            exclusion)
    #         result_with_filepath, downloads = self.file.get_fileref_and_generate_dl_list(
    #             docs, "_ed_attachment")
    #         res = search.process_data_derived_from_mongodb(
    #             result_with_filepath)
    #
    #     # 実行
    #     raw_json = dumps(res, ensure_ascii=False, indent=4)
    #     encoded_json = raw_json.encode('utf-8')
    #     json_tree_file_name = 'json_tree'
    #
    #     with tempfile.TemporaryDirectory() as tmp:
    #         p = Path(tmp)
    #         zip_filepath = self.file.zipped_contents_path(downloads,
    #                                                  json_tree_file_name,
    #                                                  encoded_json, p)
    #         # 解凍
    #         shutil.unpack_archive(zip_filepath, tmp)
    #         # 添付ファイルを読み込んでテキストの中身を抽出
    #         path_lists = [tmp, str(list(downloads.keys())[0]), name]
    #         with open(os.path.join(*path_lists)) as f:
    #             attached_s = f.read()
    #         path_lists = [tmp, json_tree_file_name + '.json']
    #         json_file_path = os.path.join(*path_lists)
    #         with open(json_file_path) as jf:
    #             j_data = json.load(jf)
    #         # 外に出すデータを組み立て(zip内部のjsonデータ、上記の添付ファイルデータ)
    #         actual_data = {'json_data': j_data, 'attached_data': attached_s}
    #     # テスト
    #     self.assertDictEqual(res, actual_data['json_data'])
    #     self.assertEqual(test_var, actual_data['attached_data'])


    def test_get_fileref_and_generate_dl_list(self):

        if not self.db_server_connect:
            return

        # DBにdocsと添付ファイルを挿入する
        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            name = 'file_dl_list.txt'
            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            input_filename = os.path.splitext(os.path.basename(name))[0]
            for filename in self.make_txt_files(tmp_dl_dir, input_filename):
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))
            # docをDBに入れる
            child_oid = ObjectId()
            doc = {'name': 'test1',
                   self.config.file: files_oid,
                   self.config.child: DBRef('child_col', child_oid)}
            insert_result = self.testdb[
                'test__get_fileref_and_generate_dl_list'].insert_one(doc)
            # ドキュメントを取得する
            res = self.testdb[
                'test__get_fileref_and_generate_dl_list'].find_one(
                {'_id': insert_result.inserted_id})
            # 実行
            new_docs, dl_list = self.file.get_fileref_and_generate_dl_list(res,
                                                                           '_ed_attachment')
            # 正常系 docsを比較
            expected = {'_id': insert_result.inserted_id, 'name': 'test1',
                        '_ed_attachment': [
                            str(insert_result.inserted_id) + '/' + name],
                        '_ed_child': DBRef('child_col', child_oid)}
            actual = new_docs
            self.assertDictEqual(expected, actual)

            # 正常系 DL処理用の辞書を比較
            expected = {insert_result.inserted_id: files_oid}
            actual = dl_list
            self.assertDictEqual(expected, actual)

    def test_generate_zip_filename(self):

        # 正常系
        # ファイル名を指定
        filename = 'testFileName'
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        expected = name + filename + '.zip'
        actual = self.file.generate_zip_filename(filename)
        self.assertEqual(expected, actual)

        # ファイル名指定なし
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        expected = name + '.zip'
        actual = self.file.generate_zip_filename()
        self.assertEqual(expected, actual)

        # 正常系
        # 文字列以外
        filename = 2545
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        filename_str = str(filename)
        expected = name + filename_str + '.zip'
        actual = self.file.generate_zip_filename(filename)
        self.assertEqual(expected, actual)

    def test_grid_out(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='file_dl', qty=2)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            test_vars = {}
            for filename in files:
                with filename.open('rb') as f:
                    content = f.read()
                    # file_obj = gzip.compress(content, compresslevel=6)
                    # files_oid.append(
                    #     self.fs.put(file_obj, filename=filename.name,
                    #                 compress='gzip'))
                    files_oid.append(
                        self.fs.put(content, filename=filename.name))
                    test_vars.update({filename.name: content.decode()})

        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            path = Path(tmp_dl_dir)

            grid_out_result = self.file._grid_out(files_oid, tmp_dl_dir)
            # 実行結果
            self.assertTrue(grid_out_result)

            # DB送信前のファイル名＆ファイルの中身と、ダウンロード後のファイル名&ファイルの中身を比較する
            expected = {}
            for dl_file in path.glob('*.txt'):
                with dl_file.open() as f:
                    expected.update({dl_file.name: f.read()})
            self.assertDictEqual(test_vars, expected)

    def test_upload(self):
        if not self.db_server_connect:
            return

        # refの場合
        # refデータ入力
        ref_json = {
            "structure_1": [
                {
                    "position": "top",
                    "username": "ryde",
                    "structure_2": [
                        {
                            "maker": "Ferrari",
                            "carname": "F355",
                            "power": 380,
                            "float_val": 4453.456
                        },
                        {
                            "maker": "HONDA",
                            "carname": "NSX",
                            "power": 280,
                            "float_val": 321.56,
                            "start_date": {"#date": "1981-04-23"},
                            "structure_3_1": [
                                {
                                    "filename": "test1.txt",
                                    "name": "添付ファイル1"
                                },
                                {
                                    "filename": "test2.txt",
                                    "name": "添付ファイル2"
                                }
                            ],
                            "structure_3_2": [
                                {
                                    "filename": "test3.txt",
                                    "name": "添付ファイル3",
                                    "structure_4": {
                                        "filename": "test4.txt",
                                        "name": "添付ファイル4",
                                        "structure_5": [
                                            {
                                                "url": "example2.com",
                                                "name": "テストURL2"
                                            },
                                            {
                                                "url": "example3.com",
                                                "name": "テストURL3"
                                            }
                                        ]
                                    },
                                    "structure_6": [
                                        {
                                            "url": "example_x.com",
                                            "name": "テストURL_x"
                                        },
                                        {
                                            "url": "example_y.com",
                                            "name": "テストURL_y"
                                        }
                                    ],
                                    "structure_5": {
                                        "url": "example.com",
                                        "name": "テストURL1",
                                        "structure_5": {
                                            "url": "example4.com",
                                            "name": "テストURL4"
                                        }
                                    }
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        convert = Convert()
        converted_edman = convert.dict_to_edman(ref_json, mode='ref')
        insert_result = self.db.insert(converted_edman)

        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='insert_file_ref_test',
                                        qty=2)
            expected = []
            for file_path in files:
                with file_path.open() as f:
                    expected.append((file_path.name, f.read()))

            # メソッド実行
            collection = 'structure_5'
            oid = [i[collection][0] for i in insert_result if collection in i][
                0]
            p_files = tuple([(i, True) for i in files])
            insert_file_result = self.file.upload(collection, oid,
                                                  p_files,
                                                  'ref')

            # DBからデータ取得
            query = {'_id': oid}
            result = self.testdb[collection].find_one(query)

            # 取得したデータからgridfsのファイルを取得
            actual = []
            self.fs = gridfs.GridFS(self.testdb)
            for file_oid in result[self.config.file]:
                data = self.fs.get(file_oid)
                d = data.read()
                if hasattr(data, 'compress') and data.compress == 'gzip':
                    d = gzip.decompress(d)
                actual.append((data.filename, d.decode()))

            # テスト
            self.assertTrue(insert_file_result)
            self.assertListEqual(sorted(actual), sorted(expected))

        # embの場合
        data = {
            "position": "top",
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "float_val": 4453.456
                },
                {
                    "maker": "HONDA",
                    "carname": "NSX",
                    "power": 280,
                    "float_val": 321.56,
                    "structure_3_1": [
                        {
                            "filename": "test1.txt",
                            "name": "添付ファイル1"
                        },
                        {
                            "filename": "test2.txt",
                            "name": "添付ファイル2"
                        }
                    ],
                    "structure_3_2": [
                        {
                            "filename": "test3.txt",
                            "name": "添付ファイル3",
                            "structure_4": {
                                "filename": "test4.txt",
                                "name": "添付ファイル4",
                                "structure_5": [
                                    {
                                        "url": "example2.com",
                                        "name": "テストURL2"
                                    },
                                    {
                                        "url": "example3.com",
                                        "name": "テストURL3"
                                    }
                                ]
                            },
                            "structure_6": [
                                {
                                    "url": "example_x.com",
                                    "name": "テストURL_x"
                                },
                                {
                                    "url": "example_y.com",
                                    "name": "テストURL_y"
                                }
                            ],
                            "structure_5": {
                                "url": "example.com",
                                "name": "テストURL1",
                                "structure_5": {
                                    "url": "example4.com",
                                    "name": "テストURL4"
                                }
                            }
                        }
                    ]
                }
            ]
        }
        insert_result = self.testdb['structure_emb'].insert_one(data)

        with tempfile.TemporaryDirectory() as tmp_dir:
            files_obj = self.make_txt_files(tmp_dir, name='emb_test', qty=2)
            expected = []
            for file_path in files_obj:
                with file_path.open() as f:
                    expected.append(f.read())

            collection = 'structure_emb'
            oid = insert_result.inserted_id
            query = ['structure_2', '1']

            files_obj2 = tuple([(i, True) for i in files_obj])
            # メソッド実行
            insert_file_emb_result = self.file.upload(
                collection, oid, files_obj2, 'emb', query)

            # ドキュメントをfindして出す
            result = self.testdb[collection].find_one({'_id': oid})

            # gridfsからデータ取得
            out_data = []
            self.fs = gridfs.GridFS(self.testdb)
            for oid in result['structure_2'][1][self.config.file]:
                data = self.fs.get(oid)
                f_data = data.read()
                if hasattr(data, 'compress') and data.compress == 'gzip':
                    f_data = gzip.decompress(f_data)

                out_data.append(f_data.decode())

            # DBデータとファイルのデータに差異はないか
            self.assertListEqual(sorted(expected), sorted(out_data))

            # メソッド成功時のフラグ
            self.assertTrue(insert_file_emb_result)

    def test_grid_in(self):
        if not self.db_server_connect:
            return

        # 正常系
        with tempfile.TemporaryDirectory() as tmp_dir:
            sample_files = self.make_txt_files(tmp_dir, name='grid_in_test',
                                               qty=2)
            compress_settings = [False, False]
            td = tuple(
                [(p, b) for (p, b) in zip(sample_files, compress_settings)])

            self.fs = gridfs.GridFS(self.testdb)
            actual = []
            for oid in self.file.grid_in(td):
                data = self.fs.get(oid)
                # if data.compress == 'gzip':
                #     f_data = gzip.decompress(data.read()).decode()
                #     b_data = True
                # else:
                #     f_data = data.read().decode()
                #     b_data = False

                f_data = data.read().decode()
                b_data = False
                actual.append([data.filename, f_data, b_data])

            expected = []
            for file_path, compress in zip(sample_files, compress_settings):
                with file_path.open() as f:
                    expected.append([file_path.name, f.read(), compress])

            self.assertListEqual(sorted(actual), sorted(expected))



import configparser
import random
import string
import copy
from pathlib import Path
from unittest import TestCase
from pymongo import errors, MongoClient
from edman import DB
from edman.exceptions import EdmanDbConnectError


class TestMultiUser(TestCase):
    db_server_connect = False
    test_ini = []
    client = None
    connections = []

    @classmethod
    def setUpClass(cls):
        # 設定読み込み
        settings = configparser.ConfigParser()
        settings.read(Path.cwd() / 'ini' / 'test_db.ini')
        cls.test_ini = dict(settings.items('DB'))
        cls.test_ini['port'] = int(cls.test_ini['port'])

        # DB作成のため、pymongoから接続
        cls.client = MongoClient(cls.test_ini['host'], cls.test_ini['port'])

        # 接続確認
        try:
            cls.client.admin.command('hello')
            cls.db_server_connect = True
            print('Use DB.')
        except errors.ConnectionFailure:
            print('Do not use DB.')

        # 作成するユーザ及びDBの数
        users_count = 4
        test_account = []

        # 人数分のユーザ名とパスワードをランダムで作成
        for i in range(users_count):
            tmp = string.digits + string.ascii_lowercase + string.ascii_uppercase
            tmp_name = ''.join([random.choice(tmp) for i in range(8)])
            tmp_pass = ''.join([random.choice(tmp) for i in range(8)])
            test_account.append(
                {
                    'username': tmp_name,
                    'password': tmp_pass,
                    'db_name': tmp_pass,
                    'db_auth': tmp_pass,
                }
            )

        if cls.db_server_connect:
            # adminで認証
            cls.client = MongoClient(
                username=cls.test_ini['admin_user'],
                password=cls.test_ini['admin_password'])

            connections = []
            for account in test_account:
                # DB作成
                cls.client[account['db_name']].command(
                    "createUser",
                    account['username'],
                    pwd=account['password'],
                    roles=[
                        {
                            'role': 'dbOwner',
                            'db': account['db_name'],
                        },
                    ],
                )

                # edmanのDB接続オブジェクト作成
                con = {
                    'host': cls.test_ini['host'],
                    'port': cls.test_ini['port'],
                    'user': account['username'],
                    'password': account['password'],
                    'database': account['db_name'],
                    'options': [f"authSource={account['db_auth']}"]
                }
                connections.append(con)
            cls.connections = connections

    @classmethod
    def tearDownClass(cls):
        # アカウント情報をもとにDBとユーザを削除
        if cls.db_server_connect:
            for connection in cls.connections:
                # cls.clientはpymongo経由でDB削除
                cls.client.drop_database(connection['database'])
                # cls.testdb.dbはedman側の接続オブジェクト経由でユーザ(自分自身)の削除
                db = DB(connection)
                testdb = db.get_db
                testdb.command("dropUser", connection['user'])

    def test_admin_insert(self):
        if not self.db_server_connect:
            return

        # 管理者権限で各ユーザのDBに入力
        connections = copy.deepcopy(self.connections)
        test_results = {}
        for connection in connections:
            connection['options'] = ['authSource=admin']
            connection['user'] = 'admin'
            connection['password'] = 'admin'
            db = DB(connection)
            insert_data = [
                {'test_admin_insert': {'database': connection['database']}}
            ]
            result = db.insert(insert_data)
            result_oid = result[0]['test_admin_insert'][0]
            # print(result_oid)  # debug
            test_results.update({connection['database']: result_oid})
        # print(test_results)  # debug

        # 各ユーザがDBからデータを取り出してテスト
        for connection in self.connections:
            db = DB(connection)
            result_doc = db.doc('test_admin_insert',
                                test_results[connection['database']],
                                None)
            # print(result_doc['database'], connection['database'])  # debug
            with self.subTest(connection=connection):
                self.assertEqual(result_doc['database'],
                                 connection['database'])

    def test_auth(self):
        if not self.db_server_connect:
            return

        # 自分のアカウントを利用して、他人のデータベースに接続する(失敗になる)

        dbnames = [connection['database'] for connection in self.connections]

        for connection in self.connections:
            # 自分以外のデータベース名のリストを作成
            tmp_dbnames = copy.deepcopy(dbnames)
            if connection['database'] in tmp_dbnames:
                tmp_dbnames.remove(connection['database'])

            # 他人のデータベースに接続を試みる
            for dbname in tmp_dbnames:
                tmp_connection = copy.deepcopy(connection)
                tmp_connection['database'] = dbname
                with self.assertRaises(EdmanDbConnectError):
                    _ = DB(tmp_connection)



from unittest import TestCase
from datetime import datetime
from bson import ObjectId, errors
import dateutil.parser
from edman import Config
from edman.utils import Utils


class TestUtils(TestCase):

    def setUp(self):
        self.config = Config()
        self.parent = self.config.parent
        self.child = self.config.child
        self.date = self.config.date
        self.file = self.config.file

    def test__item_literal_check(self):
        # 正常系 リスト内が全てリテラルデータ
        data = [1, 2, 3]
        self.assertTrue(Utils.item_literal_check(data))

        # 正常系 リスト内にオブジェクトを含むリテラルデータ
        data = [1, 2, ObjectId()]
        self.assertTrue(Utils.item_literal_check(data))

        # 正常系 リスト内に辞書
        data = [1, 2, {'d': 'bb'}]
        self.assertFalse(Utils.item_literal_check(data))

        # 正常系 リスト内にリスト
        data = [1, 2, ['1', 2]]
        self.assertFalse(Utils.item_literal_check(data))

        # 正常系 入力が辞書
        data = {'d': '34'}
        self.assertFalse(Utils.item_literal_check(data))

    def test_doc_traverse(self):
        # 正常系1 対象のキーを削除する
        doc = {
            'a': '1',
            'b': [
                {
                    'd': {
                        'e': '4',
                        'f': [ObjectId(), ObjectId(), ObjectId()],
                        'g': '5'
                    }
                },
                {
                    'c': '3'
                }
            ]
        }
        del_keys = ['f']
        query = ['b', '0', 'd']

        def delete(doc, keys):
            for key in keys:
                if key in doc:
                    del doc[key]

        actual = Utils.doc_traverse(doc, del_keys, query, delete)
        expected = {
            'a': '1',
            'b': [
                {
                    'd': {
                        'e': '4',
                        'g': '5'
                    }
                },
                {
                    'c': '3'
                }
            ]
        }
        self.assertDictEqual(actual, expected)

        # 例外のテスト1 クエリのインデックスの不備
        query = ['bbb', '2', 'eee', '2']
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1', 'hhh': 'iii'}
                ]}
            ]
        }
        oids = [ObjectId(), ObjectId(), ObjectId()]
        with self.assertRaises(IndexError):
            _ = Utils.doc_traverse(doc, oids, query, delete)

        # 例外のテスト2 クエリの指定ミス
        query = ['bbb', '2', 'eee', 'xxx']
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1', 'hhh': 'iii'}
                ]}
            ]
        }
        oids = [ObjectId(), ObjectId(), ObjectId()]
        with self.assertRaises(ValueError):
            _ = Utils.doc_traverse(doc, oids, query, delete)

    def test_conv_objectid(self):

        # 正常系 oidの場合
        oid = ObjectId()
        actual = Utils.conv_objectid(oid)
        self.assertIsInstance(actual, ObjectId)
        self.assertEqual(oid, actual)

        # 正常系 文字列の場合
        oid = ObjectId()
        actual = Utils.conv_objectid(str(oid))
        self.assertIsInstance(actual, ObjectId)
        self.assertEqual(oid, actual)

        # 異常系 oidにならない文字列
        oid = str(ObjectId())
        oid = oid[:-1]
        with self.assertRaises(errors.InvalidId) as cm:
            _ = Utils.conv_objectid(oid)

    def test__to_datetime(self):
        # datetime正常
        input_list = ['2018/11/20', '2018/11/20 13:48', '2018/01/01 00:00:00']
        for s in input_list:
            with self.subTest(s=s):
                actual = Utils.to_datetime(s)
                self.assertIsInstance(actual, datetime)

        # 入力値が文字列だがdatetime変換できない場合、または入力値が文字列以外
        input_list = [20181120, 201811201348, 20200101000000, '8月12日', None]
        for s in input_list:
            with self.subTest(s=s):
                actual = Utils.to_datetime(s)
                self.assertIsInstance(actual, str)

    def test__query_check(self):

        # 正常系
        query = ['bbb', '2', 'eee', '0', 'fff']
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1'}
                ]}
            ]
        }
        actual = Utils.query_check(query, doc)
        self.assertIsInstance(actual, bool)
        self.assertTrue(actual)

        # 異常系 間違ったクエリ
        query = ['bbb', '2', 'eee', '1', 'fff']  # インデックスの指定ミスを想定
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1'}
                ]}
            ]
        }
        actual = Utils.query_check(query, doc)
        self.assertIsInstance(actual, bool)
        self.assertFalse(actual)

    def test_item_delete(self):
        # 正常系
        doc = {
            self.parent: ObjectId(),
            self.child: [ObjectId(), ObjectId()],
            self.file: [ObjectId(), ObjectId()],
            'param': 'OK'
        }
        actual = Utils.item_delete(
            doc, ('_id', self.parent, self.child, self.file))
        expected = {'param': 'OK'}
        self.assertDictEqual(actual, expected)

    def test_child_combine(self):
        # データ構造のテスト
        test_data = [
            [
                {'collection_A': {'name': 'NSX'}},
                {'collection_A': {'name': 'F355'}},
                {'collection_B': {'power': 280}}
            ]
        ]
        actual = [i for i in Utils.child_combine(test_data)]
        self.assertIsInstance(actual, list)
        self.assertEqual(2, len(actual[0]['collection_A']))

    def test_field_name_check(self):

        illegals = [None, '', '$aa', '.aa']
        for i in illegals:
            with self.subTest(i=i):
                actual = Utils.field_name_check(i)
                self.assertFalse(actual)

        # 文字列以外の方は文字列に変換される
        actual = Utils.field_name_check(455)
        self.assertTrue(actual)

    def test_collection_name_check(self):

        illegals = [None, '', '$aaa', 'aaa$b', 'system.aaa', '#aaa', '@aaa']
        for i in illegals:
            with self.subTest(i=i):
                actual = Utils.collection_name_check(i)
                self.assertFalse(actual)

        # 文字列以外の方は文字列に変換される
        actual = Utils.collection_name_check(345)
        self.assertTrue(actual)

    def test_type_cast_conv(self):
        # 正常系 変換テスト
        input_l = ['str', 'int', 'float', 'bool', 'datetime']
        expected = [str, int, float, bool, dateutil.parser.parse]
        actual = [Utils.type_cast_conv(i) for i in input_l]
        self.assertEqual(expected, actual)

        # 正常系 str変更テスト
        input_l = ['str', '12']
        expected = [str, str]
        actual = [Utils.type_cast_conv(i) for i in input_l]
        self.assertEqual(expected, actual)



import os
import shutil
import copy
from tempfile import TemporaryDirectory, NamedTemporaryFile
import datetime
import zipfile
from typing import Union, Tuple, Iterator, List, Any
from pathlib import Path
import gridfs
from gridfs.errors import NoFile, GridFSError
import jmespath
from bson import ObjectId
from edman.utils import Utils
from edman.exceptions import (EdmanFormatError, EdmanDbProcessError,
                              EdmanInternalError)
from edman import Config


class File:
    """
    ファイル取扱クラス
    """

    def __init__(self, db=None) -> None:

        if db is not None:
            self.db = db
            self.fs = gridfs.GridFS(self.db)
        self.file_ref = Config.file
        self.comp_level = Config.gzip_compress_level

    @staticmethod
    def file_gen(files: Tuple[Path]) -> Iterator:
        """
        ファイルタプルからファイルを取り出すジェネレータ

        :param tuple files: 中身はPathオブジェクト
        :return: ファイル名と内容(str)のタプル
        :rtype: tuple
        """
        for file in files:
            try:
                with file.open('rb') as f:
                    fp = f.read()
            except IOError:
                raise
            yield file.name, fp

    def delete(self, delete_oid: ObjectId, collection: str,
               oid: Union[ObjectId, str], structure: str, query=None) -> bool:
        """
        該当のoidをファイルリファレンスから削除し、GridFSからファイルを削除

        :param ObjectId delete_oid:
        :param str collection:
        :param str oid:
        :param str structure:
        :param query:
        :type query: list or None
        :return:
        :rtype: bool
        """
        oid = Utils.conv_objectid(oid)
        if structure not in ['ref', 'emb']:
            raise EdmanFormatError('structureはrefまたはembの指定が必要です')

        # ドキュメント存在確認&コレクション存在確認&対象ドキュメント取得
        if (doc := self.db[collection].find_one({'_id': oid})) is None:
            raise EdmanInternalError('対象のコレクション、またはドキュメントが存在しません')

        # ファイルリスト取得
        files_list = self.get_file_ref(doc, structure, query)
        if len(files_list) == 0:
            raise EdmanDbProcessError('ファイルが存在しません')

        # リファレンスデータを編集
        # 何らかの原因で重複があった場合を避けるため一度setにする
        files_list = list(set(files_list))
        files_list.remove(delete_oid)

        # ドキュメントを新しいファイルリファレンスに置き換える
        try:
            if structure == 'ref':
                new_doc = self.file_list_replace(doc, files_list)
            else:
                new_doc = Utils.doc_traverse(doc, files_list, query,
                                             self.file_list_replace)
        except Exception:
            raise
        replace_result = self.db[collection].replace_one({'_id': oid}, new_doc)
        # fsから該当ファイルを削除
        if replace_result.modified_count:
            self.fs_delete([delete_oid])

        # ファイルが削除されればOK
        return False if self.fs.exists(delete_oid) else True

    def fs_delete(self, oids: list) -> None:
        """
        fsからファイル削除

        :param list oids:
        :return:
        """
        for oid in oids:
            if self.fs.exists(oid):
                self.fs.delete(oid)

    def get_file_ref(self, doc: dict, structure: str, query=None) -> list:
        """
        ファイルリファレンス情報を取得

        :param dict doc:
        :param str structure:
        :param query:
        :type: list or None
        :return: files_list
        :rtype: list
        """
        if structure == 'emb' and query is None:
            raise EdmanFormatError('embにはクエリが必要です')
        if structure != 'emb' and structure != 'ref':
            raise EdmanFormatError('構造の選択はembまたはrefが必要です')

        if structure == 'ref':
            files_list = doc.get(self.file_ref, [])
        else:
            if not Utils.query_check(query, doc):
                EdmanFormatError('対象のドキュメントに対してクエリーが一致しません.')
            # docから対象クエリを利用してファイルのリストを取得
            # deepcopyを使用しないとなぜか子のスコープのqueryがクリヤーされる
            query_c = copy.deepcopy(query)
            try:
                files_list = self._get_emb_files_list(doc, query_c)
            except Exception:
                raise

        return files_list

    def get_file_names(self, collection: str, oid: Union[ObjectId, str],
                       structure: str, query=None) -> dict:
        """
        ファイル一覧を取得
        ファイルが存在しなければ空の辞書を返す

        :param str collection:
        :param str oid:
        :param str structure:
        :param query: embの時だけ必要. refの時はNone
        :type query: list or None
        :return: result
        :rtype: dict
        """
        oid = Utils.conv_objectid(oid)
        result = {}

        # ドキュメント存在確認&コレクション存在確認&対象ドキュメント取得
        if (doc := self.db[collection].find_one({'_id': oid})) is None:
            raise EdmanDbProcessError(
                f'ドキュメントまたはコレクションが存在しません oid:{oid} collection:{collection}')

        # gridfsからファイル名を取り出す
        for file_oid in self.get_file_ref(doc, structure, query):
            try:
                fs_out = self.fs.get(file_oid)
            except gridfs.errors.NoFile:
                continue
            else:
                result.update({file_oid: fs_out.filename})
        return result

    def download(self, file_oid: list[ObjectId],
                 path: Union[str, Path]) -> bool:
        """
        Gridfsからデータをダウンロードし、ファイルに保存

        :param list file_oid:
        :param path:
        :type path: str or Path
        :return: result
        :rtype: bool
        """
        # 定型的な前処理があればここに追加する
        return self._grid_out(file_oid, path)

    def _grid_out(self, file_oid_list: List[ObjectId],
                  path: Union[str, Path]) -> bool:
        """
        Gridfsからデータを取得し、ファイルに保存
        複数のファイルを指定すると、複数のファイルが作成される

        :param list file_oid_list:
        :param path:
        :type path: str or Path
        :return: result
        :rtype: bool
        """

        # パスがstrならpathlibにする
        p = Path(path) if isinstance(path, str) else path

        # パスが正しいか検証
        if not p.exists():
            raise FileNotFoundError
        # ファイルが存在するか検証
        if False in map(self.fs.exists, file_oid_list):
            raise EdmanDbProcessError('指定のファイルはDBに存在しません')

        # ダウンロード処理
        results = []
        for file_oid in file_oid_list:
            fs_out = self.fs.get(file_oid)
            save_path = p / fs_out.filename
            try:
                with save_path.open('wb') as f:
                    tmp = fs_out.read()
                    f.write(tmp)
                    f.flush()
                    os.fsync(f.fileno())
            except IOError:
                raise
            results.append(save_path.exists())

        return all(results)

    def upload(self, collection: str, oid: Union[ObjectId, str],
               file_path: Tuple[Tuple[Any, bool]], structure: str,
               query=None) -> bool:
        """
        ドキュメントにファイルリファレンスを追加する
        ファイルのインサート処理なども行う
        :param str collection:
        :param oid:
        :type oid: ObjectId or str
        :param tuple file_path:ドキュメントに添付する単数or複数のファイルパス
        :param str structure:
        :param query:
        :type query: list or None
        :return:
        :rtype: bool
        """
        oid = Utils.conv_objectid(oid)
        if structure not in ['ref', 'emb']:
            raise EdmanFormatError('構造はrefかembが必要です')

        # ドキュメント存在確認&対象ドキュメント取得
        doc = self.db[collection].find_one({'_id': oid})
        if doc is None:
            raise EdmanInternalError('対象のドキュメントが存在しません')
        if structure == 'emb':
            # クエリーがドキュメントのキーとして存在するかチェック
            if not Utils.query_check(query, doc):
                raise EdmanFormatError('対象のドキュメントに対してクエリーが一致しません.')

        # ファイルのインサート
        inserted_file_oids = self.grid_in(file_path)
        if structure == 'ref':
            new_doc = self.file_list_attachment(doc, inserted_file_oids)
        else:
            try:
                new_doc = Utils.doc_traverse(doc, inserted_file_oids, query,
                                             self.file_list_attachment)
            except Exception:
                raise

        # ドキュメント差し替え
        replace_result = self.db[collection].replace_one({'_id': oid}, new_doc)
        if replace_result.modified_count == 1:
            result = True
        else:  # 差し替えができなかった時は添付ファイルは削除
            self.fs_delete(inserted_file_oids)
            result = False

        return result

    def grid_in(self, files: Tuple[Tuple[Any, bool]]) -> list[Any]:
        """
        Gridfsへ複数のデータをアップロード

        :param tuple files:
        :return: inserted
        :rtype: list
        """
        inserted = []
        for file, compress in files:
            try:
                with file.open('rb') as f:
                    fb = f.read()
                    metadata = {'filename': os.path.basename(f.name)}
            except (IOError, OSError) as e:
                raise EdmanDbProcessError(e)
            try:
                inserted.append(self.fs.put(fb, **metadata))
            except GridFSError as e:
                raise EdmanDbProcessError(e)
        return inserted

    def file_list_attachment(self, doc: dict,
                             files_oid: List[ObjectId]) -> dict:
        """
        辞書データにファイルのoidを挿入する
        docにself.file_refがあれば、追加する処理
        oidが重複していないものだけ追加
        ファイルが同じでも別のoidが与えられていれば追加される

        :param dict doc:
        :param list files_oid: ObjectIdのリスト
        :return: doc
        :rtype: dict
        """
        if self.file_ref in doc:
            doc[self.file_ref].extend(files_oid)
            files_oid = sorted(list(set(doc[self.file_ref])))
        # self.file_refがなければ作成してfiles_oidを値として更新
        if files_oid:
            doc.update({self.file_ref: files_oid})
        return doc

    def file_list_replace(self, doc: dict, files_oid: list) -> dict:
        """
        ドキュメントのファイルリファレンスを入力されたリストに置き換える
        もし空リストならファイルリファレンス自体を削除する
        すでにファイルリファレンスデータが存在していることを前提としているため、
        docにファイルリファレンスデータが無かった場合は例外を発生する

        :param dict doc:
        :param list files_oid:
        :return: doc
        :rtype: dict
        """
        if self.file_ref in doc:
            if files_oid:
                doc[self.file_ref] = files_oid
            else:
                del doc[self.file_ref]
        else:
            raise ValueError(f'{self.file_ref}がないか削除された可能性があります')
        return doc

    def _get_emb_files_list(self, doc: dict, query: list) -> list:
        """
        ドキュメントからファイルリファレンスのリストを取得する

        :param dict doc:
        :param list query:
        :return:
        :rtype:list
        """
        s = Utils.generate_jms_query(query)
        s += '.' + self.file_ref + '[]'
        return jmespath.search(s, doc)

    @staticmethod
    def generate_zip_filename(filename=None) -> str:
        """
        '%Y%m%d%H%M%S'.zipのファイル名を生成
        任意の文字列を指定すると, '%Y%m%d%H%M%S' + 任意の文字列 + .zipを生成

        :param any filename: strにキャストされる
        :return:
        :rtype: str
        """
        filename = filename or None
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        if filename is not None:
            name = name + str(filename)
        return name + '.zip'

    def zipped_contents(self, downloads: dict, json_tree_file_name: str,
                        encoded_json: bytes) -> str:
        """
        jsonと添付ファイルを含むzipファイルを生成
        zipファイル内部にjson_tree_file_name.jsonのjsonファイルを含む
        添付ファイルがなく、jsonファイルだけ取得したい場合はzipped_jsonを利用

        :param dict downloads:
        :param str json_tree_file_name:
        :param bytes encoded_json:
        :rtype: str
        :return:
        """

        zip_suffix = '.zip'
        json_suffix = '.json'

        with TemporaryDirectory() as tmpdir:
            dir_path_list = [os.path.join(tmpdir, str(doc_oid)) for
                             doc_oid in downloads]

            for dir_path in dir_path_list:
                os.mkdir(dir_path)

            for file_refs, dir_path in zip([i for i in downloads.values()],
                                           dir_path_list):
                for file_ref in file_refs:
                    try:
                        content = self.fs.get(file_ref)
                    except NoFile:
                        raise EdmanDbProcessError('指定の関連ファイルが存在しません')
                    except GridFSError:
                        raise
                    else:
                        filepath = os.path.join(dir_path, content.name)
                    try:
                        with open(filepath, 'wb') as f:
                            f.write(content.read())
                    except (FileNotFoundError, IOError):
                        EdmanInternalError('ファイルを保存することが出来ませんでした')
                    except GridFSError:
                        raise
            try:
                # jsonファイルとして一時ディレクトリに保存
                with open(os.path.join(tmpdir,
                                       json_tree_file_name + json_suffix),
                          'wb') as f:
                    f.write(encoded_json)
            except (FileNotFoundError, IOError):
                EdmanInternalError('ファイルを保存することが出来ませんでした')
            try:
                # 最終的にDLするzipファイルを作成
                with NamedTemporaryFile() as fp:
                    zip_filepath = shutil.make_archive(fp.name, zip_suffix[1:],
                                                       tmpdir)
            except Exception:
                raise
        return zip_filepath

    # def zipped_contents_path(self, downloads: dict, json_tree_file_name: str,
    #                     encoded_json: bytes, p: Path) -> str:
    #     """
    #     jsonと添付ファイルを含むzipファイルを生成
    #     zipファイル内部にjson_tree_file_name.jsonのjsonファイルを含む
    #     添付ファイルがなく、jsonファイルだけ取得したい場合はzipped_jsonを利用
    #
    #     :param dict downloads:
    #     :param str json_tree_file_name:
    #     :param bytes encoded_json:
    #     :param Path p:
    #     :rtype: str
    #     :return:
    #     """
    #
    #     json_suffix = '.json'
    #     dir_path_list = [p/str(doc_oid) for doc_oid in downloads]
    #
    #     for file_refs, dir_path in zip([i for i in downloads.values()],
    #                                    dir_path_list):
    #         os.mkdir(dir_path)
    #         for file_ref in file_refs:
    #             content = self.fs.get(file_ref)
    #             filepath = dir_path / content.name
    #             with open(filepath, 'wb') as f:
    #                 f.write(content.read())
    #
    #         json_dir = dir_path / json_tree_file_name + json_suffix
    #         # jsonファイルとして一時ディレクトリに保存
    #         with json_dir.open('wb') as f:
    #             f.write(encoded_json)
    #
    #     # 最終的にDLするzipファイルを作成
    #     zip_filepath = shutil.make_archive('dl', 'zip', p)
    #
    #     return zip_filepath


    @staticmethod
    def zipped_json(encoded_json: bytes, json_tree_file_name: str) -> str:
        """
        jsonファイルとzipファイルを名前付きテンポラリとして生成し、パスを生成
        zipファイル内部にjson_tree_file_name.jsonのjsonファイルを含む
        添付ファイルがなく、jsonファイルだけ取得したい場合に使用する

        :param bytes encoded_json: json文字列としてダンプしたものを指定の文字コードでバイト列に変換したもの
        :param str json_tree_file_name: zipファイル内に配置するjsonファイルの名前
        :return:
        :rtype:str
        """

        json_suffix = '.json'
        zip_suffix = '.zip'
        try:
            # jsonファイルとして一時ファイルに保存
            with NamedTemporaryFile(suffix=json_suffix,
                                    delete=False) as fp:
                filepath = fp.name
                fp.write(encoded_json)
        except Exception:
            raise
        try:
            # jsonファイルをzipで圧縮して一時ファイルに保存
            with NamedTemporaryFile(suffix=zip_suffix,
                                    delete=False) as fp:
                with zipfile.ZipFile(fp.name, 'w',
                                     zipfile.ZIP_DEFLATED) as archive:
                    zip_filepath = fp.name
                    archive.write(filepath,
                                  arcname=json_tree_file_name + json_suffix)
        except Exception:
            raise
        return zip_filepath

    @staticmethod
    def zipped_json_path(encoded_json: bytes, json_tree_file_name: str,
                     p: Path) -> Path:
        """
        zipファイル内部にjson_tree_file_name.jsonのjsonファイルを含む
        添付ファイルがなく、jsonファイルだけ取得したい場合に使用する

        :param bytes encoded_json: json文字列としてダンプしたものを指定の文字コードでバイト列に変換したもの
        :param str json_tree_file_name: zipファイル内に配置するjsonファイルの名前
        :param Path p: ファイルを保存するディレクトリのパス
        :return:zip_filepath
        :rtype:Path
        """
        zip_filename = json_tree_file_name + '.zip'
        filename = json_tree_file_name + '.json'
        jsonfile = p / filename
        zip_filepath = p / zip_filename

        try:
            with jsonfile.open('wb') as f:
                f.write(encoded_json)
        except Exception:
            raise
        try:
            with zipfile.ZipFile(zip_filepath, 'w', zipfile.ZIP_DEFLATED) as f:
                f.write(jsonfile, arcname=filename)
        except Exception:
            raise
        return zip_filepath

    def get_fileref_and_generate_dl_list(self, docs: dict,
                                         attach_key: str) -> tuple[dict, dict]:
        """
        json出力用の辞書内のファイルリファレンスをファイルパス文字列リストに置き換える
            例:
            {"_ed_file":[ObjectId('file_oid_1'),ObjectId('file_oid_2')]}
            ↓
            {"_ed_attachment":["document_oid/sample.jpg","document_oid/sample2.jpg"]}
        同時にダウンロード処理用の辞書を作成する
            {ObjectId('document_oid'):[ObjectId('file_oid_1'),ObjectId('file_oid_2')]}

        :param dict docs:
        :param str attach_key:
        :return:
        :rtype: tuple
        """
        dl_list = {}

        def recursive(data: dict, doc_oid=None):
            c_docs = {}
            for key, value in data.items():
                if isinstance(data[key], dict):
                    # ドキュメントのoidを取得
                    if '_id' in data[key]:
                        doc_oid = data[key]['_id']
                    c_docs.update({key: recursive(data[key], doc_oid)})
                elif isinstance(value, list) and Utils.item_literal_check(
                        value):
                    if Config.file in key:
                        # ファイルリファレンスオブジェクト取得
                        try:
                            file_out = [self.fs.get(i) for i in value]
                        except NoFile:
                            raise EdmanDbProcessError('指定の関連ファイルが存在しません')
                        except GridFSError:
                            raise
                        # ファイルパス生成
                        tmp = {
                            attach_key: [
                                str(data['_id']) + '/' + i.filename for i
                                in file_out]
                        }
                        dl_list.update({data['_id']: value})
                    else:
                        tmp = {key: value}
                    c_docs.update(tmp)
                elif isinstance(data[key], list):
                    c_docs.update(
                        {key: [recursive(item, doc_oid) for item in
                               data[key]]})
                else:
                    c_docs.update({key: value})
            return c_docs

        new_docs = recursive(docs)
        return new_docs, dl_list



from unittest import TestCase
from datetime import datetime
from bson import ObjectId, DBRef
from edman import Config, Convert


class TestConvert(TestCase):

    def setUp(self):
        self.convert = Convert()
        self.config = Config()
        self.parent = self.config.parent
        self.child = self.config.child
        self.date = self.config.date

    def test__get_child_reference(self):

        # すでにparentが作られていた場合
        # (この関数後の処理にDBRefオブジェクトが0個の時の処理が書かれているため0個でもOK)
        test_data = {self.parent: {'_id': ObjectId()}}
        actual = self.convert._get_child_reference(test_data)
        self.assertEqual(0, len(list(actual.values())[0]))

        # テスト辞書データの中身が辞書の場合
        test_data = {'test_collection': {'_id': ObjectId()}}
        actual = self.convert._get_child_reference(test_data)
        self.assertIsInstance(actual, dict)
        self.assertEqual(self.config.child, list(actual.keys())[0])
        self.assertEqual(1, len(list(actual.values())[0]))

        # テスト辞書データの中身がリストの場合
        test_data = {
            'test_collection': [
                {'_id': ObjectId()},
                {'_id': ObjectId()}
            ]
        }
        actual = self.convert._get_child_reference(test_data)
        self.assertEqual(2, len(list(actual.values())[0]))

    def test__convert_datetime(self):
        # データ構造のテスト
        test_data = {'start_date': {'#date': '1981-04-23'}}
        actual = self.convert._convert_datetime(test_data)
        self.assertIsInstance(actual, dict)
        self.assertIsInstance(list(actual.values())[0], datetime)

    def test__list_organize(self):
        # データ構造のテスト
        test_data = [
            {
                'honda': [
                    {'name': 'NSX', 'power': '280'}
                ]
            },
            {
                'honda': [
                    {'name': 'S800', 'first model year': '1966'}
                ]
            }
        ]
        actual = self.convert._list_organize(test_data)
        self.assertIsInstance(actual, list)
        self.assertEqual('honda', list(actual[0].keys())[0])
        self.assertEqual(2, len(list(actual[0].values())[0]))
        for i in list(actual[0].values())[0]:
            with self.subTest(i=i):
                self.assertIsInstance(i, dict)

    def test__list_intercept_hook(self):

        data = {
            'parent_data': 'data1',
            'delete_coll': {'data': '1', 'data2': '2'},
            'child': [
                DBRef('test_collection', ObjectId()),
                DBRef('test_collection', ObjectId())]}
        actual = self.convert._list_intercept_hook('test_coll', data)

        # 子要素のデータ(この場合はdata['delete_coll'])が削除されているか
        self.assertIsInstance(actual, dict)
        for k, v in actual.items():
            with self.subTest(v=v):
                self.assertNotIn('delete_coll', v[0])

    def test__attached_oid(self):
        # 辞書内辞書の場合
        test_data = {
            'collection': {
                'car_name': 'NSX',
                'Purchase year': '1993'
            }
        }
        actual = self.convert._attached_oid(test_data)
        self.assertIsInstance(actual, dict)
        self.assertTrue(True if '_id' in list(actual.values())[0] else False)
        self.assertIsInstance(list(actual.values())[0]['_id'], ObjectId)

        # 辞書内リストの場合
        test_data = {
            'collection': [
                {
                    'car_name': 'NSX',
                    'Purchase year': '1993'
                }
            ]
        }
        actual = self.convert._attached_oid(test_data)
        self.assertTrue(
            True if '_id' in list(actual.values())[0][0] else False)
        self.assertIsInstance(list(actual.values())[0][0]['_id'], ObjectId)

    def test__date_replace(self):
        list_data = [{self.date: '2019-02-28'},
                     {self.date: '2019-03-01 13:56:28'},
                     1,
                     'text']
        expected = [datetime(2019, 2, 28),
                    datetime(2019, 3, 1, 13, 56, 28),
                    1,
                    'text']
        actual = self.convert._date_replace(list_data)
        self.assertListEqual(expected, actual)

    def test_pullout_key(self):

        data = {
            'beamtime': {
                'username': 'guest',
                'expInfo': [
                    {
                        'date': '2019-08-02',
                        'position': {'a': '1', 'b': '2'},
                        'file': {'name': 'sample.jpg'},
                        'process': {
                            'cc': '33',
                            'file': {
                                'name': 'machine.log',
                                'file': {'name': 'machine2.log'}
                            }
                        }
                    },
                    {
                        'date': '2019-08-03',
                        'position': {'a': '11', 'b': '22'},
                        'file': {'name': 'process.jpg'},
                        'process': {
                            'cc': '44',
                            'file': {
                                'name': 'machine3.log',
                                'file': {'name': 'machine4.log'}
                            }
                        }
                    },
                ]}
        }
        key = 'expInfo'
        actual = self.convert.pullout_key(data, key)

        expected = {
            'expInfo': [
                {
                    'date': '2019-08-02',
                    'position': {'a': '1', 'b': '2'},
                    'file': {'name': 'sample.jpg'},
                    'process': {
                        'cc': '33',
                        'file': {
                            'name': 'machine.log',
                            'file': {'name': 'machine2.log'}
                        }
                    }
                },
                {
                    'date': '2019-08-03',
                    'position': {'a': '11', 'b': '22'},
                    'file': {'name': 'process.jpg'},
                    'process': {
                        'cc': '44',
                        'file': {
                            'name': 'machine3.log',
                            'file': {'name': 'machine4.log'}
                        }
                    }
                },
            ]}
        self.assertIsInstance(actual, dict)
        self.assertDictEqual(expected, actual)

        # 別のデータ(この場合は1番目のprocessを拾ってくる)
        data2 = {
            'beamtime': {
                'username': 'guest',
                'expInfo': [
                    {
                        'date': '2019-08-02',
                        'position': {'a': '1', 'b': '2'},
                        'file': {'name': 'sample.jpg'},
                        'process': {
                            'cc': '33',
                            'file': {
                                'name': 'machine.log',
                                'file': {'name': 'machine2.log'}
                            }
                        }
                    },
                    {
                        'date': '2019-08-03',
                        'position': {'a': '11', 'b': '22'},
                        'file': {'name': 'process.jpg'},
                        'process': {
                            'cc': '44',
                            'file': {
                                'name': 'machine3.log',
                                'file': {'name': 'machine4.log'}
                            }
                        }
                    },
                ]}
        }
        key = 'process'
        actual = self.convert.pullout_key(data2, key)
        expected = {
            'process': {
                'cc': '33',
                'file': {
                    'name': 'machine.log',
                    'file': {'name': 'machine2.log'}
                }
            }
        }
        self.assertDictEqual(expected, actual)

        # 指定のキーが見つからなかった時
        key = 'foobar'
        actual = self.convert.pullout_key(data, key)
        self.assertDictEqual({}, actual)

        # データが無かった場合
        data3 = {}
        key = 'expInfo'
        actual = self.convert.pullout_key(data3, key)
        self.assertDictEqual({}, actual)

    def test_exclusion_key(self):
        data = {
            'expInfo': [
                {
                    'date': '2019-08-02',
                    'position': {'a': '1', 'b': '2'},
                    'file': {'name': 'sample.jpg'},
                    'process': {
                        'cc': '33',
                        'file': {
                            'name': 'machine.log',
                            'file': {'name': 'machine2.log'}
                        }
                    }
                },
                {
                    'date': '2019-08-03',
                    'position': {'a': '11', 'b': '22'},
                    'file': {'name': 'process.jpg'},
                    'process': {
                        'cc': '44',
                        'file': {
                            'name': 'machine3.log',
                            'file': {'name': 'machine4.log'}
                        }
                    }
                },
            ]}

        exclusion = ('position',)
        actual = self.convert.exclusion_key(data, exclusion)
        expected = {
            'expInfo': [
                {
                    'date': '2019-08-02',
                    'file': {'name': 'sample.jpg'},
                    'process': {
                        'cc': '33',
                        'file': {
                            'name': 'machine.log',
                            'file': {'name': 'machine2.log'}
                        }
                    }
                },
                {
                    'date': '2019-08-03',
                    'file': {'name': 'process.jpg'},
                    'process': {
                        'cc': '44',
                        'file': {
                            'name': 'machine3.log',
                            'file': {'name': 'machine4.log'}
                        }
                    }
                },
            ]}
        self.assertIsInstance(actual, dict)
        self.assertDictEqual(expected, actual)

        # 指定のキーが見つからなかった時
        exclusion = ('foobar',)
        actual = self.convert.exclusion_key(data, exclusion)
        self.assertDictEqual(data, actual)

    def test__ref(self):
        pass
        # テスト用jsonの読み込み
        # with open('./test_json_files/ref_test_premo.json') as f:
        #     jsondict = json.load(f)
        #
        # actual = self.convert._ref(jsondict)
        #
        # with open('./test_json_files/ref_test_premo_result.json', 'w') as f:
        #     f.write(dumps(actual, ensure_ascii=False, indent=4))
        # TODO あとでテストデータを追加し、assertできるようにする

    def test__emb(self):
        pass
        # with open('./test_json_files/ref_test_premo.json') as f:
        #     jsondict = json.load(f)
        #
        # actual = self.convert._emb(jsondict)
        # with open('./test_json_files/emb_test_premo_result.json', 'w') as f:
        #     f.write(dumps(actual, ensure_ascii=False, indent=4))
        # TODO あとでテストデータを追加し、assertできるようにする

    #
    # def test_dict_to_edman(self):
    #     # 中身は他のメソッドなのでテストはパス
    #     pass



import configparser
import copy
from pathlib import Path
from datetime import datetime
from unittest import TestCase
import dateutil.parser
from pymongo import errors as py_errors, MongoClient
from bson import ObjectId, DBRef, errors
from edman import Config, DB, Search


class TestSearch(TestCase):
    db_server_connect = False
    test_ini = []
    client = None

    @classmethod
    def setUpClass(cls):
        # 設定読み込み
        settings = configparser.ConfigParser()
        settings.read(Path.cwd() / 'ini' / 'test_db.ini')
        cls.test_ini = dict(settings.items('DB'))
        cls.test_ini['port'] = int(cls.test_ini['port'])

        # DB作成のため、pymongoから接続
        cls.client = MongoClient(cls.test_ini['host'], cls.test_ini['port'])

        # 接続確認
        try:
            cls.client.admin.command('hello')
            cls.db_server_connect = True
            print('Use DB.')
        except py_errors.ConnectionFailure:
            print('Do not use DB.')

        if cls.db_server_connect:
            # adminで認証
            cls.client = MongoClient(
                username=cls.test_ini['admin_user'],
                password=cls.test_ini['admin_password'])
            # DB作成
            cls.client[cls.test_ini['db']].command(
                "createUser",
                cls.test_ini['user'],
                pwd=cls.test_ini['password'],
                roles=[
                    {
                        'role': 'dbOwner',
                        'db': cls.test_ini['db'],
                    },
                ],
            )
            # edmanのDB接続オブジェクト作成
            con = {
                'host': cls.test_ini['host'],
                'port': cls.test_ini['port'],
                'user': cls.test_ini['user'],
                'password': cls.test_ini['password'],
                'database': cls.test_ini['db'],
                'options': [f"authSource={cls.test_ini['db']}"]
            }
            db = DB(con)
            cls.testdb = db.get_db
            cls.search = Search(db)
        else:
            cls.search = Search()

    @classmethod
    def tearDownClass(cls):
        if cls.db_server_connect:
            # cls.clientはpymongo経由でDB削除
            # cls.testdb.dbはedman側の接続オブジェクト経由でユーザ(自分自身)の削除
            cls.client.drop_database(cls.test_ini['db'])
            cls.testdb.command("dropUser", cls.test_ini['user'])

    def setUp(self):
        self.config = Config()
        self.parent = self.config.parent
        self.child = self.config.child
        self.date = self.config.date
        self.file = self.config.file

    def tearDown(self):
        if self.db_server_connect:
            # システムログ以外のコレクションを削除
            collections_all = self.testdb.list_collection_names()
            log_coll = 'system.profile'
            if log_coll in collections_all:
                collections_all.remove(log_coll)
            for collection in collections_all:
                self.testdb.drop_collection(collection)

    def test__merge_parent(self):

        # データ構造のテスト
        # 本来は親子関係なのでchild項目も存在するが、テストに関係ないので省略
        dummy_id = ObjectId()
        dummy_id2 = ObjectId()
        parent_id = ObjectId()
        parent_data = {
            'parent_col': {
                '_id': dummy_id,
                'a': 'aa',
                'parent_col2': {
                    '_id': dummy_id2,
                    'b': 'bb',
                    self.parent: DBRef('parent_col', dummy_id),
                    'parent_col3': {
                        '_id': parent_id,
                        'c': 'cc',
                        self.parent: DBRef('parent_col2', dummy_id2)
                    }
                }
            }
        }
        self_data = {
            'self_col': {
                'car_name': 'F355',
                self.parent: DBRef('parent_col3', parent_id)
            }
        }
        actual = self.search._merge_parent(parent_data, self_data)
        expected = \
            actual['parent_col']['parent_col2']['parent_col3']['self_col'][
                'car_name']

        self.assertIsInstance(actual, dict)
        self.assertEqual('F355', expected)

    def test__objectid_replacement(self):
        # 正常系
        query = {'_id': '5bf4f3ce9663380fc50d6dbd'}
        actual = self.search._objectid_replacement(query)
        self.assertIsInstance(actual['_id'], ObjectId)

        # 異常系
        query = {'_id': 'dragon'}
        with self.assertRaises(errors.InvalidId) as cm:
            _ = self.search._objectid_replacement(query)

    def test__get_self(self):
        if not self.db_server_connect:
            return

        # テストデータをDBに挿入
        data = {'test_data': 'test'}
        test_collection = 'get_self_test'
        db = self.client[self.test_ini['db']]
        insert_result = db[test_collection].insert_one(data)

        # メソッドを実行してデータを取ってくる dictや構造
        # 作成したデータと取得したデータの差異
        query = {'_id': insert_result.inserted_id}
        actual = self.search._get_self(query, test_collection)
        self.assertIsInstance(actual, dict)
        self.assertEqual(sorted(list(data.keys())),
                         sorted(list(actual[test_collection].keys())))

        # 存在しないデータの時はNoneを返す
        query = {'_id': insert_result.inserted_id, 'test_data': 'foo'}
        actual = self.search._get_self(query, test_collection)
        self.assertIsNone(actual)

    def test__get_parent(self):
        if not self.db_server_connect:
            return

        # テストデータをDBに挿入
        db = self.client[self.test_ini['db']]
        data1_id = ObjectId()
        data2_id = ObjectId()
        data3_id = ObjectId()
        parent_coll = 'parent_coll'
        parent2_coll = 'parent2_coll'
        self_coll = 'self_coll'
        data1 = {
            '_id': data1_id,
            'data1': 'test',
            self.child: [DBRef(parent2_coll, data2_id)]
        }
        _ = db[parent_coll].insert_one(data1)
        data2 = {
            '_id': data2_id,
            'data2': 'test',
            self.parent: DBRef(parent_coll, data1_id),
            self.child: [DBRef(self_coll, data3_id)]
        }
        _ = db[parent2_coll].insert_one(data2)
        data3 = {
            '_id': data3_id,
            'data3': 'test',
            self.parent: DBRef(parent2_coll, data2_id)
        }
        _ = db[self_coll].insert_one(data3)

        self_result = {self_coll: data3}
        actual = self.search._get_parent(self_result, depth=2)
        self.assertIsInstance(actual, dict)

        for k, v in actual.items():
            # 構造のチェック
            self.assertEqual(k, parent_coll)
            self.assertEqual(v['_id'], data1['_id'])
            self.assertIsInstance(v[parent2_coll], dict)
            self.assertEqual(v[parent2_coll]['_id'], data2_id)

            # 親データ内のネストチェック
            self.assertEqual(v[self.child][0].id, v[parent2_coll]['_id'])
            self.assertEqual(v['_id'], v[parent2_coll][self.parent].id)

            # selfとparentのネストチェック
            self.assertEqual(v[parent2_coll][self.child][0].id, data3['_id'])
            self.assertEqual(v[parent2_coll]['_id'], data3[self.parent].id)

    def test__build_to_doc_parent(self):
        # データ構造のテスト
        # parentに近い方から順番に並んでいる(一番最後がrootまたはrootに近い方)
        parent_data_list = [
            {
                'parent_2': {
                    'car_name': 'STORATOS'
                }
            },
            {
                'parent_1': {
                    'program': 'python'
                }
            }
        ]
        actual = self.search._build_to_doc_parent(parent_data_list)
        self.assertIsInstance(actual, dict)
        expected = actual['parent_1']['parent_2']['car_name']
        self.assertEqual('STORATOS', expected)

    def test_process_data_derived_from_mongodb(self):

        # 正常系
        data = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }

        def _item_literal_check(list_child):
            result = True
            if isinstance(list_child, dict):
                result = False
            if isinstance(list_child, list):
                for j in list_child:
                    if isinstance(j, dict) or isinstance(j, list):
                        result = False
                        break
            return result

        # リファレンス削除チェック用メソッド
        def detect_ref(d):
            if self.parent == d or self.child == d:
                raise ValueError('parent or child, was not deleted')

        # 日付変換チェック用メソッド
        def detect_date(d):
            if isinstance(d, datetime):
                raise ValueError('datetime was not deleted')

        # データを回すだけの簡単なお仕事
        def rec(data: dict):
            for key, val in data.items():
                if isinstance(data[key], list) and _item_literal_check(
                        data[key]):
                    for i in data[key]:
                        detect_date(i)
                elif isinstance(data[key], list):
                    for item in data[key]:
                        detect_ref(item)
                        rec(item)
                elif isinstance(data[key], dict):
                    detect_ref(data[key])
                    rec(data[key])

                # 型変換
                else:
                    detect_date(val)

        actual = self.search.process_data_derived_from_mongodb(data)
        self.assertIsInstance(actual, dict)
        self.assertIsNone(rec(actual))

        # 正常系 refsの指定 resultに_id, ed_parent, _ed_childを残す
        data = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        expected = copy.deepcopy(data)
        del expected['coll1']['coll2']['coll3'][0][self.file]
        refs = ['_id', self.parent, self.child]
        actual = self.search.process_data_derived_from_mongodb(data,
                                                               exclusion=refs)
        self.assertDictEqual(actual, expected)

        # 正常系 refsに設定するが空リストの場合
        data = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        expected = copy.deepcopy(data)
        del expected['coll1']['_id']
        del expected['coll1'][self.child]
        del expected['coll1']['coll2']['_id']
        del expected['coll1']['coll2'][self.parent]
        del expected['coll1']['coll2'][self.child]
        del expected['coll1']['coll2']['coll3'][0]['_id']
        del expected['coll1']['coll2']['coll3'][0][self.parent]
        del expected['coll1']['coll2']['coll3'][0][self.child]
        del expected['coll1']['coll2']['coll3'][0][self.file]
        del expected['coll1']['coll2']['coll3'][1]['_id']
        del expected['coll1']['coll2']['coll3'][1][self.parent]
        del expected['coll1']['coll2']['coll3'][1][self.child]
        actual = self.search.process_data_derived_from_mongodb(data,
                                                               exclusion=[])
        self.assertDictEqual(actual, expected)

        # 異常系 リファレンス系以外のキーを指定した場合
        data_e1 = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        with self.assertRaises(ValueError):
            _ = self.search.process_data_derived_from_mongodb(data_e1,
                                                              exclusion=[
                                                                  'test'])

        # 異常系 exclusionにリストとNone以外の値が入力された場合
        data_e2 = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        with self.assertRaises(ValueError):
            _ = self.search.process_data_derived_from_mongodb(data_e2,
                                                              exclusion=(1, 2))

    def test__format_datetime(self):
        # 正常系
        data = ['2018/11/22', '2018/11/22 11:45:23', '2019-01-01 00:00:00']
        for i in map(dateutil.parser.parse, data):
            # 入力値と出力値がdatetime形式として一致しているか確認
            with self.subTest(i=i):
                actual = self.search._format_datetime(i)
                self.assertEqual(i, dateutil.parser.parse(actual[self.date]))

            # 構造のテスト
            self.assertIsInstance(actual, dict)
            self.assertIsInstance(actual[self.date], str)
            self.assertEqual(self.date, list(actual.keys())[0])


        # 正常系 階層指定 中身がdb.find()なので割愛
        # all_docs = self.search.get_documents(1, doc_col, doc_id,
        #                                      parent_depth=0, child_depth=1)
        # print('all_docs:', all_docs)
        # 正常系 単一のドキュメント 中身がdb.find()なので割愛
        # all_docs = self.search.get_documents(3, doc_col, doc_id,
        #                                      parent_depth=0, child_depth=0)
        # print('all_docs:', all_docs)

    # def test_find(self):
    #     # 中身は別のメソッドが中心なのでテストは割愛
    #     pass
    #
    # def test__self_data_select(self):
    #     # 画面上の選択処理なので、テストは割愛
    #     pass
    #
